#include "camera.hpp"
#include "ffmpeg/avcodec.hpp"
#if !defined(NDEBUG)
#	include <iostream>
#	include <iomanip>
#endif
#include <opencv2/flann.hpp>
#include <climits>
#include <fstream>

constexpr const auto MjpegFormat = mgs::fourcc_to_int("MJPG");

#if !defined(NDEBUG)
namespace mgs {
std::ostream& operator<< (std::ostream& stream, const FrameFormat& value) {
	stream << value.width << 'x' << value.height << ' ' <<
		static_cast<char>((value.format >> 0 * CHAR_BIT) & 0xFF) <<
		static_cast<char>((value.format >> 1 * CHAR_BIT) & 0xFF) <<
		static_cast<char>((value.format >> 2 * CHAR_BIT) & 0xFF) <<
		static_cast<char>((value.format >> 3 * CHAR_BIT) & 0xFF);
	return stream;
}
} //namespace mgs
#endif

namespace {
#if !defined(NDEBUG)
void print_supported_frame_formats (const mgs::Camera& cam) {
	const auto curr_frame_size = cam.frame_size();

	std::cout << "Supported frame formats:\n";
	int count = 0;
	for (const auto& size : cam.supported_frame_sizes()) {
		std::cout << std::setw(2) << ++count << " - " << size;

		if (curr_frame_size == size)
			std::cout << " *";
		std::cout << '\n';
	}
}
#endif
} //unnamed namespace

int main() {
	try {
		mgs::Camera cam("/dev/video0", {}, false);
		cam.set_frame_size({1280, 720, MjpegFormat});

		mgs::ffmpeg::AVCodec avcodec{::AVCodecID::AV_CODEC_ID_MJPEG};
		auto cam_frame_format = cam.frame_size();
		auto decoder{avcodec.make_decoder(cam_frame_format.width, cam_frame_format.height, ::AVPixelFormat::AV_PIX_FMT_RGB24)};

		print_supported_frame_formats(cam);
		cam.start_streaming();

		std::vector<std::uint8_t> my_frame;
		auto data = cam.read_frame(my_frame);
		std::cout << "Read frame of " << data.size() << " bytesfrom camera\n";

		std::ofstream file_wr("temp.jpg", std::ios::binary | std::ios::out);
		file_wr.write(reinterpret_cast<const char*>(data.data()), data.size());

		std::vector<std::uint8_t> buff;
		buff.resize(cam_frame_format.width * 3 * cam_frame_format.height);
		const auto out_data_size = decoder.decode(data.data(), data.size(), buff.data(), buff.size());
		std::cout << "Allocated buffer size: " << buff.size() << '\n';
		std::cout << "Decoded buffer size: " << out_data_size << '\n';

		std::ofstream rgb_file_wr("rgb.data", std::ios::binary | std::ios::out);
		rgb_file_wr.write(reinterpret_cast<const char*>(buff.data()), buff.size());
	}
	catch (const std::runtime_error& err) {
		std::cerr << "Unhandled exception: " << err.what() << '\n';
		return 1;
	}

	std::cout << "Program done\n";
	return 0;
}
